import datetime
import pandas as pd

from .build_data_frame import build_data_frame
from .run_analysis import analyze_df
from .build_summary import build_summary


def take_action(row, strategy, columns):
    """
    Params:
        row: data row to operate on
        strategy: dictionary of logic and how to impliment it
        columns: list of data points describing th row
    Returns:
        boolean, True if row meets the criteria of given strategy,
        False if otherwise
    """
    results = []

    df_col_map = dict(zip(columns, row))
    for each in strategy:
        check1 = each[0] if type(each[0]) == int else df_col_map[each[0]]
        check2 = each[2] if type(each[2]) == int else df_col_map[each[2]]

        if each[1] == ">":
            results.append(bool(check1 > check2))
        if each[1] == "<":
            results.append(bool(check1 < check2))
        if each[1] == "=":
            results.append(bool(check1 == check2))
        if each[1] == "!=":
            results.append(bool(check1 != check2))

    return all(results)


def determine_action(frame, strategy, df_col_map):
    """
    Params:
        frame: current row of the dataframe
        strategy: object with the logic of how to trade
        df_col_map: dictionary with the column name and index
        of the dataframe
    Returns:
        string, "e" (enter), "x" (exit), "h" (hold) of what
        the strategy would do
    """

    action = "h"

    if take_action(frame, strategy["enter"], df_col_map):
        action = "e"

    if take_action(frame, strategy["exit"], df_col_map):
        action = "x"

    return action


def run_backtest(strategy, ohlcv_path="", df=None):
    """
    Params:
        strategy: dict, required, object containing the logic to test
        ohlcv_path: string or list, required, where to find the csv file of the ohlcv data
        df
    Returns: dictionary
            summary is a dict summary of the performace of backtest
            df is a pandas dataframe object used in the backtest
    """

    start = datetime.datetime.utcnow()

    if not df:
        df = build_data_frame(strategy, ohlcv_path)

    flagged_enter, flagged_exit, strategy = get_flagged_logiz(strategy)

    strategy["base_balance"] = strategy.get("base_balance", 1000)
    strategy["exit_on_end"] = strategy.get("exit_on_end", True)

    df = process_dataframe(df, strategy)

    if flagged_enter or flagged_exit:
        strategy["enter"].extend(flagged_enter)
        strategy["exit"].extend(flagged_exit)

        df = process_dataframe(df, strategy)

    summary = build_summary(df, start)

    return {"summary": summary, "df": df}


def get_flagged_logiz(strategy):
    """
    removes logiz that need to be processed after the initial run
    """
    to_flag = ["aux_perc_change"]
    flagged_exit_idx = []
    flagged_enter_idx = []

    flagged_enter = []
    flagged_exit = []

    for idx, ex in enumerate(strategy["exit"]):
        if ex[0] in to_flag or ex[2] in to_flag:
            flagged_exit_idx.append(idx)

    for e_ex in flagged_exit_idx:
        flagged_exit.append(strategy["exit"].pop(e_ex))

    for idx, en in enumerate(strategy["enter"]):
        if en[0] in to_flag or en[2] in to_flag:
            flagged_enter_idx.append(idx)

    for e_en in flagged_enter:
        flagged_enter.append(strategy["enter"].pop(e_en))

    return flagged_enter, flagged_exit, strategy


def process_dataframe(df, strategy):
    """
    Processes the frame and adds the resultant rows
    """
    df["actions"] = [
        determine_action(frame, strategy, list(df.columns)) for frame in df.values
    ]
    base, aux, smooth_base = analyze_df(df, strategy)

    if len(base) != len(df.index):
        new_row = pd.DataFrame(
            df[-1:].values,
            index=[df.index[-1] + pd.Timedelta(minutes=1)],
            columns=df.columns,
        )
        df = df.append(new_row)

    df["base_balance"] = base
    df["aux_balance"] = aux
    df["smooth_base"] = smooth_base

    df["aux_perc_change"] = df["smooth_base"].pct_change() * 100
    df["aux_change"] = df["smooth_base"].diff()

    return df



# flake8: noqa
from .build_data_frame import build_data_frame
from .run_analysis import enter_trade, exit_trade, analyze_df
from .run_backtest import run_backtest, determine_action, take_action
from .cli import main



import pandas as pd
from finta import TA
import os

MIN = 1
HOUR = 60
DAY = 1440


def build_data_frame(strategy, ohlcv_path):
    """
    Params:
        ohlcv_path: string, absolute path of where to find the data file
        strategy: dict, assembles and calulates all the data, designed by user
    """
    df = load_basic_df_from_csv(ohlcv_path)

    indicators = strategy.get("indicators", [])

    df = apply_indicators_to_dataframe(df, indicators)

    s_chart_period = strategy.get("chart_period", 1)
    chart_period = determine_chart_period(s_chart_period)

    start_time = strategy.get("start")
    stop_time = strategy.get("stop")

    df = apply_charting_to_df(df, chart_period, start_time, stop_time)

    if df.empty:
        raise Exception("Dataframe is empty. Check the start and end dates")

    return df


def apply_charting_to_df(df, chart_period, start_time, stop_time):
    """
    Args
        df: dataframe with data loaded
    Returns:
        a sorted dataframe with the appropriate time frames
    """

    # print("hmm:", df.index[0])
    time_unit = detect_time_unit(df.index[0])

    df["datetime"] = pd.to_datetime(df.index, unit=time_unit)
    df.set_index(["datetime"], inplace=True)

    df = df.iloc[::chart_period, :]

    if start_time and stop_time:
        df = df[start_time:stop_time]  # noqa
    elif start_time and not stop_time:
        df = df[start_time:]  # noqa
    elif not start_time and stop_time:
        df = df[:stop_time]  # noqa

    return df


def apply_indicators_to_dataframe(df, indicators):
    for ind in indicators:
        func = ind.get("func")
        field_name = ind.get("name")
        if ind.get("args"):
            args = ind.get("args", [])
            df[field_name] = indicator_map[func](df, *args)
        else:
            df[field_name] = indicator_map[func](df)

    return df


def load_basic_df_from_csv(ohlcv_path):
    """
    ohlcv_path: string or list of file paths
    """
    if type(ohlcv_path) == str:
        if not os.path.isfile(ohlcv_path):
            raise Exception(f"File not found: {ohlcv_path}")

        df = pd.read_csv(ohlcv_path, parse_dates=True)

    if type(ohlcv_path) == list:
        for idx, new_file_path in enumerate(ohlcv_path):
            if not os.path.isfile(new_file_path):
                raise Exception(f"File not found: {new_file_path}")

            if idx == 0:
                df = pd.read_csv(new_file_path, parse_dates=True)
            else:

                df = df.merge(pd.read_csv(new_file_path, parse_dates=True))

    df.set_index(["date"], inplace=True)
    return df


def determine_chart_period(chart_period):
    multiplyer = MIN

    chart_period = str(chart_period)

    if chart_period.isnumeric():
        clean_chart_period = int(chart_period)

    clean_chart_period = chart_period
    if type(chart_period) == str:
        if "m" in chart_period:
            multiplyer = MIN
            replace_char = "m"

        if "h" in chart_period:
            multiplyer = HOUR
            replace_char = "h"

        if "d" in chart_period:
            replace_char = "d"
            multiplyer = DAY
        if not chart_period.isnumeric():
            clean_chart_period = int(chart_period.replace(replace_char, ""))
        else:
            clean_chart_period = int(chart_period)

    return int(clean_chart_period * multiplyer)


def detect_time_unit(timestamp):
    """
    Params:
        timestamp, usually a np.int64
    Returns: string of "s" if the string is 10 characters long, "ms" if otherwise
    """
    if len(str(timestamp)) == 10:
        return "s"

    return "ms"


indicator_map = {
    "ta.sma": TA.SMA,
    "ta.smm": TA.SMM,
    "ta.ssma": TA.SSMA,
    "ta.ema": TA.EMA,
    "ta.dema": TA.DEMA,
    "ta.tema": TA.TEMA,
    "ta.trima": TA.TRIMA,
    "ta.vama": TA.VAMA,
    "ta.er": TA.ER,
    "ta.kama": TA.KAMA,
    "ta.zlema": TA.ZLEMA,
    "ta.wma": TA.WMA,
    "ta.hma": TA.HMA,
    "ta.evwma": TA.EVWMA,
    "ta.vwap": TA.VWAP,
    "ta.smma": TA.SMMA,
    "ta.macd": TA.MACD,
    "ta.ppo": TA.PPO,
    "ta.vw_macd": TA.VW_MACD,
    "ta.ev_macd": TA.EV_MACD,
    "ta.mom": TA.MOM,
    "ta.roc": TA.ROC,
    "ta.rsi": TA.RSI,
    "ta.ift_rsi": TA.IFT_RSI,
    "ta.tr": TA.TR,
    "ta.atr": TA.ATR,
    "ta.sar": TA.SAR,
    "ta.bbands": TA.BBANDS,
    "ta.bbwidth": TA.BBWIDTH,
    "ta.percent_b": TA.PERCENT_B,
    "ta.kc": TA.KC,
    "ta.do": TA.DO,
    "ta.dmi": TA.DMI,
    "ta.adx": TA.ADX,
    "ta.pivot": TA.PIVOT,
    "ta.pivot_fib": TA.PIVOT_FIB,
    "ta.stoch": TA.STOCH,
    "ta.stochd": TA.STOCHD,
    "ta.stochrsi": TA.STOCHRSI,
    "ta.williams": TA.WILLIAMS,
    "ta.uo": TA.UO,
    "ta.ao": TA.AO,
    "ta.mi": TA.MI,
    "ta.vortex": TA.VORTEX,
    "ta.kst": TA.KST,
    "ta.tsi": TA.TSI,
    "ta.tp": TA.TP,
    "ta.adl": TA.ADL,
    "ta.chaikin": TA.CHAIKIN,
    "ta.mfi": TA.MFI,
    "ta.obv": TA.OBV,
    "ta.wobv": TA.WOBV,
    "ta.vzo": TA.VZO,
    "ta.pzo": TA.PZO,
    "ta.efi": TA.EFI,
    "ta.cfi": TA.CFI,
    "ta.ebbp": TA.EBBP,
    "ta.emv": TA.EMV,
    "ta.cci": TA.CCI,
    "ta.copp": TA.COPP,
    "ta.basp": TA.BASP,
    "ta.baspn": TA.BASPN,
    "ta.cmo": TA.CMO,
    "ta.chandelier": TA.CHANDELIER,
    "ta.qstick": TA.QSTICK,
    "ta.tmf": TA.TMF,
    "ta.fish": TA.FISH,
    "ta.ichimoku": TA.ICHIMOKU,
    "ta.apz": TA.APZ,
    "ta.vr": TA.VR,
    "ta.sqzmi": TA.SQZMI,
    "ta.vpt": TA.VPT,
    "ta.fve": TA.FVE,
    "ta.vfi": TA.VFI,
    "ta.msd": TA.MSD,
    "ta.wto": TA.WTO,
}



import pytest
import pandas as pd
import datetime
from fast_trade.build_data_frame import detect_time_unit, determine_chart_period, load_basic_df_from_csv, apply_indicators_to_dataframe, apply_charting_to_df
from unittest import mock, TestCase


def test_detect_time_unit_s():
    mock_timestring = 1595115901734
    result = detect_time_unit(mock_timestring)

    assert result == "ms"

def test_detect_time_unit_ms():
    mock_timestring = 1595115901
    result = detect_time_unit(mock_timestring)

    assert result == "s"

def test_determine_chart_period_1_num():
    mock_chart_period = 1
    result = determine_chart_period(mock_chart_period)

    assert result == 1

def test_determine_chart_period_multi_num_1():
    mock_chart_period = 10
    result = determine_chart_period(mock_chart_period)

    assert result == 10

def test_determine_chart_period_multi_num_2():
    mock_chart_period = 14
    result = determine_chart_period(mock_chart_period)

    assert result == 14

def test_determine_chart_period_1_min_str_1():
    mock_chart_period = "1m"
    result = determine_chart_period(mock_chart_period)

    assert result == 1

def test_determine_chart_period_15_min_str_1():
    mock_chart_period = "15m"
    result = determine_chart_period(mock_chart_period)

    assert result == 15

def test_determine_chart_period_1_hour_str_1():
    mock_chart_period = "1h"
    result = determine_chart_period(mock_chart_period)

    assert result == 60

def test_determine_chart_period_4_hour_str_2():
    mock_chart_period = "4h"
    result = determine_chart_period(mock_chart_period)

    assert result == 240

def test_determine_chart_period_1_day_str_1():
    mock_chart_period = "1d"
    result = determine_chart_period(mock_chart_period)

    assert result == 1440

def test_determine_chart_period_5_day_str_1():
    mock_chart_period = "5d"
    result = determine_chart_period(mock_chart_period)

    assert result == 7200


def test_load_basic_df_from_csv_str_1():
    mock_ohlcv_path = "./test/ohlcv_data.csv.txt"

    result_df = load_basic_df_from_csv(mock_ohlcv_path)
    header = list(result_df.head())

    assert "close" in header
    assert "open" in header
    assert "high" in header
    assert "low" in header
    assert "volume" in header

    assert result_df.index.name == "date"

def test_load_basic_df_from_csv_list_1():
    mock_ohlcv_path = [ 
        "./test/ohlcv_data.csv.txt",
        "./test/extra_data.txt"
    ]

    result_df = load_basic_df_from_csv(mock_ohlcv_path)

    header = list(result_df.head())
    
    assert "close" in header
    assert "open" in header
    assert "high" in header
    assert "low" in header
    assert "volume" in header
    assert "extra_column_1" in header

    assert "date" in result_df.index.name

    expected_line = [0.01404, 0.01, 0.025, 0.01, 3117.0, 10.0]

    assert list(result_df.iloc[1]) == expected_line

def test_load_basic_df_from_csv_str_error_1():
    mock_ohlcv_path = "./test/SomeFakeNews.csv.txt"

    with pytest.raises(Exception, match=r"File not found:*"):
        load_basic_df_from_csv(mock_ohlcv_path)

def test_load_basic_df_from_csv_list_error_1():
    mock_ohlcv_path = [
        "./test/ohlcv_data.csv.txt",
        "./test/literally_fake_news.csv",
    ]

    with pytest.raises(Exception, match=r"File not found:*"):
        load_basic_df_from_csv(mock_ohlcv_path)

def test_apply_indicators_to_dataframe_1_ind():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_indicators = [
        {
            "func": "ta.sma",
            "name": "example_indicator_name",
            "args": [3]
        }
    ]

    result_df = apply_indicators_to_dataframe(mock_df, mock_indicators)

    header = list(result_df.head())

    assert "example_indicator_name" in header
    assert "FAKE_indicator_name" not in header

    assert result_df.index.name == None

def test_apply_indicators_to_dataframe_multi_ind():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_indicators = [
        {
            "func": "ta.sma",
            "name": "sma",
            "args": [19]
        },
        {
            "func": "ta.rsi",
            "name": "rsi",
            "args": []
        },
        {
            "func": "ta.uo",
            "name": "uo",
            "args": []
        }
    ]

    result_df = apply_indicators_to_dataframe(mock_df, mock_indicators)

    header = list(result_df.head())

    assert "rsi" in header
    assert "sma" in header
    assert "uo" in header

    assert result_df.index.name == None


def test_apply_charting_to_df_1():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.set_index(["date"], inplace=True)
    mock_chart_period = 2
    mock_start_time = "2018-04-17 04:00:00"
    mock_stop_time = ""

    result_df = apply_charting_to_df(mock_df, mock_chart_period, mock_start_time, mock_stop_time)

    assert (result_df.iloc[2].name - result_df.iloc[1].name).total_seconds() == 120
    assert (result_df.iloc[4].name - result_df.iloc[1].name).total_seconds() == 360

def test_apply_charting_to_df_2():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.set_index(["date"], inplace=True)
    mock_chart_period = 1
    mock_start_time = "2018-04-17 04:00:00"
    mock_stop_time = "2018-04-17 04:10:00"

    past_stop_time = datetime.datetime.strptime("2018-04-17 04:11:00", "%Y-%m-%d %H:%M:%S")    
    result_df = apply_charting_to_df(mock_df, mock_chart_period, mock_start_time, mock_stop_time)

    assert result_df.iloc[-1].name < past_stop_time

def test_apply_charting_to_df_2():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.set_index(["date"], inplace=True)
    mock_chart_period = 1
    mock_start_time = ""
    mock_stop_time = "2018-04-17 04:10:00"

    past_stop_time = "2018-04-17 04:11:00"

    result_df = apply_charting_to_df(mock_df, mock_chart_period, mock_start_time, mock_stop_time)


    past_stop_time = datetime.datetime.strptime("2018-04-17 04:11:00", "%Y-%m-%d %H:%M:%S")    
    result_df = apply_charting_to_df(mock_df, mock_chart_period, mock_start_time, mock_stop_time)

    assert result_df.iloc[-1].name < past_stop_time

def test_apply_to_charting_to_df_3():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.set_index(["date"], inplace=True)

    mock_chart_period = 1
    mock_start_time = "2018-04-17 04:03:00"
    mock_stop_time = "2018-04-17 04:10:00"

    result_df = apply_charting_to_df(mock_df, mock_chart_period, mock_start_time, mock_stop_time)

    past_stop_time = datetime.datetime.strptime("2018-04-17 04:11:00", "%Y-%m-%d %H:%M:%S")    
    result_df = apply_charting_to_df(mock_df, mock_chart_period, mock_start_time, mock_stop_time)

    assert result_df.iloc[-1].name < past_stop_time



import sys
import json

from .cli_helpers import (
    open_strat_file,
    format_all_help_text,
    save,
    create_plot,
    format_command,
)
from .run_backtest import run_backtest
import matplotlib.pyplot as plt


def parse_args(raw_args):
    """
    args: raw args after the command line

    returns: dict of args as key value pairs
    """
    arg_dict = {}
    for raw_arg in raw_args:
        arg = raw_arg.split("=")
        arg_key = arg[0].split("--").pop()
        if len(arg) > 1:
            if arg[1] == "false":
                arg_dict[arg_key] = False
            elif arg[1] == "true":
                arg_dict[arg_key] == True
                # flake8: noqa
            elif arg[1] != "false" or arg[1] != "true":
                arg_dict[arg_key] = arg[1]
        else:
            arg_dict[arg_key] = True

    return arg_dict


def main():
    if len(sys.argv) < 2:
        print(format_all_help_text())
        return

    command = sys.argv[1]

    args = parse_args(sys.argv[2:])

    if command == "backtest":
        # check for help
        if "help" in args.keys():
            print(format_command(command))
            return

        strat_obj = open_strat_file(args["strat"])
        strat_obj = {**strat_obj, **args}

        res = run_backtest(strat_obj, args["csv"])

        print(json.dumps((res["summary"]), indent=2))

        if args.get("save"):
            save(res, strat_obj)

        if args.get("plot"):
            create_plot(res["df"])

            plt.show()

        return

    if command == "help":
        print(format_all_help_text())
        return

    print("Command not found")
    print(format_all_help_text())


if __name__ == "__main__":
    main()



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="fast-trade",
    version="0.1.3",
    description="Automate and backtest on ohlcv data quickly",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jrmeier/fast-trade",
    py_modules=["fast_trade"],
    keywords=[
        "backtesting",
        "currency",
        "ta",
        "pandas",
        "finance",
        "numpy",
        "analysis",
        "technical analysis",
    ],
    author="Jed Meier",
    author_email="fast_trade@jedm.dev",
    license="MIT",
    python_requires=">=3",
    entry_points={"console_scripts": ["ft = fast_trade.cli:main"]},
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(include=["fast_trade"]),
    include_package_data=True,
    install_requires=[
        "finta==0.4.1",
        "numpy==1.18.1",
        "pandas==1.0.1",
        "python-dateutil==2.8.1",
        "pytz==2019.3",
        "six==1.14.0",
    ],
)

# setup(
#     name='yourscript',
#     version='0.1',
#     py_modules=['yourscript'],
#     install_requires=[
#         'Click',
#     ],
#     entry_points='''
#         [console_scripts]
#         yourscript=yourscript:cli
#     ''',
# )



import datetime
import pandas as pd
from .build_data_frame import detect_time_unit


def build_summary(df, perf_start_time):

    # Not yet implimented
    # Exposure [%]                            94.29
    # Buy & Hold Return [%]                  703.46
    # Max. Drawdown [%]                      -33.61
    # Avg. Drawdown [%]                       -5.68
    # Max. Drawdown Duration      689 days 00:00:00
    # Avg. Drawdown Duration       41 days 00:00:00
    # Expectancy [%]                           6.92
    # SQN                                      1.77
    # Sharpe Ratio                             0.22
    # Sortino Ratio                            0.54
    # Calmar Ratio                             0.07
    time_unit = detect_time_unit(df.index[0])
    aux = df.drop_duplicates(subset="aux_balance", keep=False, inplace=False)
    trade_perc_series = aux.aux_balance.pct_change() * 100
    trade_time_held_series = pd.to_datetime(aux.aux_balance, unit=time_unit).diff()
    # return {}

    mean_trade_time_held = trade_time_held_series.mean()
    max_trade_time_held = trade_time_held_series.max()
    min_trade_time_held = trade_time_held_series.min()
    median_time_held = trade_time_held_series.median()

    max_trade_perc = trade_perc_series.max()
    min_trade_perc = trade_perc_series.min()
    mean_trade_perc = trade_perc_series.mean()
    median_trade_perc = trade_perc_series.median()

    win_trades = trade_perc_series[trade_perc_series > 0]
    loss_trades = trade_perc_series[trade_perc_series < 0]
    total_trades = len(aux)

    try:
        win_perc = (len(win_trades) / total_trades) * 100
    except ZeroDivisionError:
        win_perc = 0

    try:
        loss_perc = (len(loss_trades) / total_trades) * 100
    except ZeroDivisionError:
        loss_perc = 0

    try:
        return_perc = (aux.iloc[-1].aux_balance / aux.iloc[0].aux_balance) * 100 - 100
    except IndexError:
        return_perc = 0
    except ZeroDivisionError:
        return_perc = 0

    equity_peak = df["aux_balance"].max()

    equity_final = df.iloc[-1]["aux_balance"]

    perf_stop_time = datetime.datetime.utcnow()
    start_date = df.index[0]
    end_date = df.index[-1]

    summary = {
        "return_perc": round(return_perc, 3),
        "median_trade_len": median_time_held.total_seconds(),
        "mean_trade_len": mean_trade_time_held.total_seconds(),
        "max_trade_held": max_trade_time_held.total_seconds(),
        "min_trade_len": min_trade_time_held.total_seconds(),
        "best_trade_perc": round(max_trade_perc, 3),
        "min_trade_perc": round(min_trade_perc, 3),
        "median_trade_perc": round(median_trade_perc, 3),
        "mean": round(mean_trade_perc, 3),
        "num_trades": len(aux),
        "win_perc": round(win_perc, 3),
        "loss_perc": round(loss_perc, 3),
        "equity_peak": float(equity_peak),
        "equity_final": float(equity_final),
        "first_tic": start_date.strftime("%Y-%m-%d %H:%M:%S"),
        "last_tic": end_date.strftime("%Y-%m-%d %H:%M:%S"),
        "total_tics": len(df.index),
        "test_duration": (perf_stop_time - perf_start_time).total_seconds(),
    }

    return summary
